from unittest import mock

from django.core.cache import cache
from django.core.files.uploadedfile import SimpleUploadedFile
from django.urls import reverse

from admin_action_tools.constants import CACHE_KEYS, CONFIRMATION_RECEIVED
from admin_action_tools.tests.helpers import AdminConfirmTestCase
from tests.factories import ItemFactory, ShopFactory
from tests.market.admin import ItemAdmin, ShoppingMallAdmin
from tests.market.models import GeneralManager, Item, ShoppingMall, Town


@mock.patch.object(ShoppingMallAdmin, "inlines", [])
class TestConfirmationCache(AdminConfirmTestCase):
    def test_simple_add(self):
        # Load the Add Item Page
        ItemAdmin.confirm_add = True
        response = self.client.get(reverse("admin:market_item_add"))

        # Should be asked for confirmation
        self.assertTrue(response.context_data.get("confirm_add"))
        self.assertIn("_confirm_add", response.rendered_content)

        # Click "Save"
        data = {
            "name": "name",
            "price": 2.0,
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_add": True,
            "_save": True,
        }
        response = self.client.post(reverse("admin:market_item_add"), data=data)

        # Should be shown confirmation page
        self._assertSubmitHtml(
            rendered_content=response.rendered_content,
            save_action="_save",
            multipart_form=True,
        )

        # Should have cached the unsaved item
        cached_item = cache.get(CACHE_KEYS["object"])
        self.assertIsNotNone(cached_item)
        self.assertIsNone(cached_item.id)
        self.assertEqual(cached_item.name, data["name"])
        self.assertEqual(cached_item.price, data["price"])
        self.assertEqual(cached_item.currency, data["currency"])

        # Should not have saved the item yet
        self.assertEqual(Item.objects.count(), 0)

        # Click "Yes, I'm Sure"
        del data["_confirm_add"]
        data[CONFIRMATION_RECEIVED] = True
        response = self.client.post(reverse("admin:market_item_add"), data=data)

        # Should have redirected to changelist
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/market/item/")

        # Should have saved item
        self.assertEqual(Item.objects.count(), 1)
        saved_item = Item.objects.all().first()
        self.assertEqual(saved_item.name, data["name"])
        self.assertEqual(saved_item.price, data["price"])
        self.assertEqual(saved_item.currency, data["currency"])

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_simple_change(self):
        item = ItemFactory(name="Not name")

        # Load the Change Item Page
        ItemAdmin.confirm_change = True
        response = self.client.get(f"/admin/market/item/{item.id}/change/")

        # Should be asked for confirmation
        self.assertTrue(response.context_data.get("confirm_change"))
        self.assertIn("_confirm_change", response.rendered_content)

        # Click "Save And Continue"
        data = {
            "id": item.id,
            "name": "name",
            "price": 2.0,
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_change": True,
            "_continue": True,
        }
        response = self.client.post(f"/admin/market/item/{item.id}/change/", data=data)

        # Should be shown confirmation page
        self._assertSubmitHtml(
            rendered_content=response.rendered_content,
            save_action="_continue",
            multipart_form=True,
        )

        # Should have cached the unsaved item
        cached_item = cache.get(CACHE_KEYS["object"])
        self.assertIsNotNone(cached_item)

        # Should not have saved the changes yet
        self.assertEqual(Item.objects.count(), 1)
        item.refresh_from_db()
        self.assertEqual(item.name, "Not name")

        # Click "Yes, I'm Sure"
        del data["_confirm_change"]
        data[CONFIRMATION_RECEIVED] = True
        response = self.client.post(f"/admin/market/item/{item.id}/change/", data=data)

        # Should not have redirected to changelist
        self.assertEqual(response.url, f"/admin/market/item/{item.id}/change/")

        # Should have saved item
        self.assertEqual(Item.objects.count(), 1)
        saved_item = Item.objects.all().first()
        self.assertEqual(saved_item.name, data["name"])
        self.assertEqual(saved_item.price, data["price"])
        self.assertEqual(saved_item.currency, data["currency"])

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_file_and_image_add(self):
        # Load the Add Item Page
        ItemAdmin.confirm_add = True
        response = self.client.get(reverse("admin:market_item_add"))

        # Should be asked for confirmation
        self.assertTrue(response.context_data.get("confirm_add"))
        self.assertIn("_confirm_add", response.rendered_content)

        # Select files
        image_path = "admin_action_tools/tests/snapshot/screenshot.png"
        f = SimpleUploadedFile(
            name="test_file.jpg",
            content=open(image_path, "rb").read(),
            content_type="image/jpeg",
        )
        i = SimpleUploadedFile(
            name="test_image.jpg",
            content=open(image_path, "rb").read(),
            content_type="image/jpeg",
        )
        # Click "Save"
        data = {
            "name": "name",
            "price": 2.0,
            "currency": Item.VALID_CURRENCIES[0][0],
            "file": f,
            "image": i,
            "_confirm_add": True,
            "_save": True,
        }
        response = self.client.post(reverse("admin:market_item_add"), data=data)

        # Should be shown confirmation page
        self._assertSubmitHtml(
            rendered_content=response.rendered_content,
            save_action="_save",
            multipart_form=True,
        )

        # Should have cached the unsaved item
        cached_item = cache.get(CACHE_KEYS["object"])
        self.assertIsNotNone(cached_item)
        self.assertIsNone(cached_item.id)
        self.assertEqual(cached_item.name, data["name"])
        self.assertEqual(cached_item.price, data["price"])
        self.assertEqual(cached_item.currency, data["currency"])
        self.assertEqual(cached_item.file, data["file"])
        self.assertEqual(cached_item.image, data["image"])

        # Should not have saved the item yet
        self.assertEqual(Item.objects.count(), 0)

        # Click "Yes, I'm Sure"
        confirmation_data = data.copy()
        del confirmation_data["_confirm_add"]
        del confirmation_data["image"]
        del confirmation_data["file"]
        confirmation_data[CONFIRMATION_RECEIVED] = True
        response = self.client.post(reverse("admin:market_item_add"), data=confirmation_data)

        # Should have redirected to changelist
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/market/item/")

        # Should have saved item
        self.assertEqual(Item.objects.count(), 1)
        saved_item = Item.objects.all().first()
        self.assertEqual(saved_item.name, data["name"])
        self.assertEqual(saved_item.price, data["price"])
        self.assertEqual(saved_item.currency, data["currency"])
        self.assertIsNotNone(saved_item.file)
        self.assertIsNotNone(saved_item.image)

        self.assertRegex(saved_item.file.name, r"test_file.*\.jpg$")
        self.assertRegex(saved_item.image.name, r"test_image.*\.jpg$")

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_file_and_image_change(self):
        item = ItemFactory(name="Not name")
        # Select files
        image_path = "admin_action_tools/tests/snapshot/screenshot.png"
        f = SimpleUploadedFile(
            name="test_file.jpg",
            content=open(image_path, "rb").read(),
            content_type="image/jpeg",
        )
        i = SimpleUploadedFile(
            name="test_image.jpg",
            content=open(image_path, "rb").read(),
            content_type="image/jpeg",
        )
        item.file = f
        item.image = i
        item.save()

        # Load the Change Item Page
        ItemAdmin.confirm_change = True
        ItemAdmin.fields = ["name", "price", "file", "image", "currency"]
        response = self.client.get(f"/admin/market/item/{item.id}/change/")

        # Should be asked for confirmation
        self.assertTrue(response.context_data.get("confirm_change"))
        self.assertIn("_confirm_change", response.rendered_content)

        # Upload new image and remove file
        i2 = SimpleUploadedFile(
            name="test_image2.jpg",
            content=open(image_path, "rb").read(),
            content_type="image/jpeg",
        )
        # Click "Save And Continue"
        data = {
            "id": item.id,
            "name": "name",
            "price": 2.0,
            "image": i2,
            "file": "",
            "file-clear": "on",
            "currency": Item.VALID_CURRENCIES[0][0],
            "_confirm_change": True,
            "_continue": True,
        }
        response = self.client.post(f"/admin/market/item/{item.id}/change/", data=data)

        # Should be shown confirmation page
        self._assertSubmitHtml(
            rendered_content=response.rendered_content,
            save_action="_continue",
            multipart_form=True,
        )

        # Should have cached the unsaved item
        cached_item = cache.get(CACHE_KEYS["object"])
        self.assertIsNotNone(cached_item)

        # Should not have saved the changes yet
        self.assertEqual(Item.objects.count(), 1)
        item.refresh_from_db()
        self.assertEqual(item.name, "Not name")
        self.assertIsNotNone(item.file)
        self.assertIsNotNone(item.image)

        # Click "Yes, I'm Sure"
        del data["_confirm_change"]
        data["image"] = ""
        data[CONFIRMATION_RECEIVED] = True
        response = self.client.post(f"/admin/market/item/{item.id}/change/", data=data)

        # Should not have redirected to changelist
        self.assertEqual(response.url, f"/admin/market/item/{item.id}/change/")

        # Should have saved item
        self.assertEqual(Item.objects.count(), 1)
        saved_item = Item.objects.all().first()
        self.assertEqual(saved_item.name, data["name"])
        self.assertEqual(saved_item.price, data["price"])
        self.assertEqual(saved_item.currency, data["currency"])
        self.assertFalse(saved_item.file)
        self.assertIsNotNone(saved_item.image)

        self.assertRegex(saved_item.image.name, r"test_image2.*\.jpg$")

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_relations_add(self):
        gm = GeneralManager.objects.create(name="gm")
        shops = [ShopFactory() for i in range(3)]
        town = Town.objects.create(name="town")

        # Load the Add ShoppingMall Page
        ShoppingMallAdmin.confirm_add = True
        response = self.client.get(reverse("admin:market_shoppingmall_add"))

        # Should be asked for confirmation
        self.assertTrue(response.context_data.get("confirm_add"))
        self.assertIn("_confirm_add", response.rendered_content)

        # Click "Save"
        data = {
            "name": "name",
            "shops": [s.id for s in shops],
            "general_manager": gm.id,
            "town": town.id,
            "_confirm_add": True,
            "_save": True,
        }
        response = self.client.post(reverse("admin:market_shoppingmall_add"), data=data)

        # Should be shown confirmation page
        self._assertManyToManyFormHtml(
            rendered_content=response.rendered_content,
            options=shops,
            selected_ids=data["shops"],
        )
        self._assertSubmitHtml(rendered_content=response.rendered_content, save_action="_save")

        # Should not have cached the unsaved object
        cached_item = cache.get(CACHE_KEYS["object"])
        self.assertIsNone(cached_item)

        # Should not have saved the object yet
        self.assertEqual(ShoppingMall.objects.count(), 0)

        # Click "Yes, I'm Sure"
        confirmation_received_data = data
        del confirmation_received_data["_confirm_add"]

        response = self.client.post(reverse("admin:market_shoppingmall_add"), data=confirmation_received_data)

        # Should have redirected to changelist
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/market/shoppingmall/")

        # Should have saved object
        self.assertEqual(ShoppingMall.objects.count(), 1)
        saved_item = ShoppingMall.objects.all().first()
        self.assertEqual(saved_item.name, data["name"])
        self.assertEqual(saved_item.general_manager, gm)
        self.assertEqual(saved_item.town, town)
        for shop in saved_item.shops.all():
            self.assertIn(shop, shops)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    def test_relation_change(self):
        gm = GeneralManager.objects.create(name="gm")
        shops = [ShopFactory() for i in range(3)]
        town = Town.objects.create(name="town")
        mall = ShoppingMall.objects.create(name="mall", general_manager=gm, town=town)
        mall.shops.set(shops)

        # new values
        gm2 = GeneralManager.objects.create(name="gm2")
        shops2 = [ShopFactory() for i in range(3)]
        town2 = Town.objects.create(name="town2")

        # Load the Change ShoppingMall Page
        ShoppingMallAdmin.confirm_change = True
        response = self.client.get(f"/admin/market/shoppingmall/{mall.id}/change/")

        # Should be asked for confirmation
        self.assertTrue(response.context_data.get("confirm_change"))
        self.assertIn("_confirm_change", response.rendered_content)

        # Click "Save"
        data = {
            "id": mall.id,
            "name": "name",
            "shops": [s.id for s in shops2],
            "general_manager": gm2.id,
            "town": town2.id,
            "_confirm_change": True,
            "_continue": True,
        }
        response = self.client.post(f"/admin/market/shoppingmall/{mall.id}/change/", data=data)

        # Should be shown confirmation page
        self._assertManyToManyFormHtml(
            rendered_content=response.rendered_content,
            options=shops,
            selected_ids=data["shops"],
        )
        self._assertSubmitHtml(rendered_content=response.rendered_content, save_action="_continue")

        # Should not have cached the unsaved obj
        cached_item = cache.get(CACHE_KEYS["object"])
        self.assertIsNone(cached_item)

        # Should not have saved changes yet
        self.assertEqual(ShoppingMall.objects.count(), 1)
        mall.refresh_from_db()
        self.assertEqual(mall.name, "mall")
        self.assertEqual(mall.general_manager, gm)
        self.assertEqual(mall.town, town)
        for shop in mall.shops.all():
            self.assertIn(shop, shops)

        # Click "Yes, I'm Sure"
        confirmation_received_data = data
        del confirmation_received_data["_confirm_change"]
        confirmation_received_data[CONFIRMATION_RECEIVED] = True

        response = self.client.post(
            f"/admin/market/shoppingmall/{mall.id}/change/",
            data=confirmation_received_data,
        )

        # Should not have redirected to changelist
        self.assertEqual(response.url, f"/admin/market/shoppingmall/{mall.id}/change/")

        # Should have saved obj
        self.assertEqual(ShoppingMall.objects.count(), 1)
        saved_item = ShoppingMall.objects.all().first()
        self.assertEqual(saved_item.name, data["name"])
        self.assertEqual(saved_item.general_manager, gm2)
        self.assertEqual(saved_item.town, town2)

        for shop in saved_item.shops.all():
            self.assertIn(shop, shops2)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))



from unittest.mock import MagicMock

from django.contrib.admin.options import IS_POPUP_VAR
from django.contrib.admin.sites import AdminSite
from django.contrib.auth.models import Permission, User
from django.urls import reverse

from admin_action_tools.constants import CONFIRM_ACTION
from admin_action_tools.tests.helpers import AdminConfirmTestCase
from tests.market.admin import ShopAdmin
from tests.market.admin.inventory_admin import InventoryAdmin
from tests.market.models import Inventory, Shop


class TestConfirmActions(AdminConfirmTestCase):
    @classmethod
    def setUpTestData(cls):
        cls.superuser = User.objects.create_superuser(  # nosec
            username="super", email="super@email.org", password="pass"
        )

    def test_get_changelist_should_not_be_affected(self):
        response = self.client.get(reverse("admin:market_shop_changelist"))
        self.assertIsNotNone(response)
        self.assertNotIn("Confirm Action", response.rendered_content)

    def test_action_without_confirmation(self):
        post_params = {
            "action": ["show_message_no_confirmation"],
            "select_across": ["0"],
            "index": ["0"],
            "_selected_action": ["3", "2", "1"],
        }
        response = self.client.post(
            reverse("admin:market_shop_changelist"),
            data=post_params,
            follow=True,  # Follow the redirect to get content
        )
        self.assertIsNotNone(response)
        self.assertEqual(response.status_code, 200)
        # Should not use confirmaiton page
        self.assertNotIn("action_confirmation", response.template_name)

        # The action was to show user a message
        self.assertIn("You selected without confirmation", response.rendered_content)

    def test_action_with_confirmation_should_show_confirmation_page(self):
        post_params = {
            "action": ["show_message"],
            "select_across": ["0"],
            "index": ["0"],
            "_selected_action": ["3", "2", "1"],
        }
        response = self.client.post(
            reverse("admin:market_shop_changelist"),
            data=post_params,
            follow=True,  # Follow the redirect to get content
        )
        self.assertIsNotNone(response)
        self.assertEqual(response.status_code, 200)
        # Should use confirmaiton page
        self.assertEqual(
            response.template_name,
            [
                "admin/market/shop/confirm_tool/action_confirmation.html",
                "admin/market/confirm_tool/action_confirmation.html",
                "admin/confirm_tool/action_confirmation.html",
            ],
        )

        # The action was to show user a message, and should not happen yet
        self.assertNotIn("You selected", response.rendered_content)

    def test_no_permissions_in_database_for_action_with_confirmation(self):
        """
        Django would not show the action in changelist action selector
        If the user doesn't have permissions, but this doesn't prevent
        user from calling post with the params to perform the action.

        If the permissions are denied because of Permission in the database,
        Django would redirect to the changelist.
        """
        # Create a user without permissions for action
        user = User.objects.create_user(  # nosec
            username="user",
            email="user@email.org",
            password="pass",
            is_active=True,
            is_staff=True,
            is_superuser=False,
        )
        # Give user permissions to ShopAdmin change, add, view but not delete
        for permission in Permission.objects.filter(codename__in=["change_shop", "view_shop", "add_shop"]):
            user.user_permissions.add(permission)

        self.client.force_login(user)

        post_params = {
            "action": ["show_message"],
            "select_across": ["0"],
            "index": ["0"],
            "_selected_action": ["3", "2", "1"],
        }
        response = self.client.post(
            reverse("admin:market_shop_changelist"),
            data=post_params,
            follow=True,  # Follow the redirect to get content
        )
        self.assertIsNotNone(response)
        self.assertEqual(response.status_code, 200)
        # Should not use confirmaiton page
        self.assertEqual(
            response.template_name,
            [
                "admin/market/shop/change_list.html",
                "admin/market/change_list.html",
                "admin/change_list.html",
            ],
        )

        # The action was to show user a message, and should not happen
        self.assertNotIn("You selected", response.rendered_content)

        # Django won't show the action as an option to you
        self.assertIn("No action selected", response.rendered_content)

    def test_no_permissions_in_code_non_superuser_for_action_with_confirmation(self):
        """
        Django would not show the action in changelist action selector
        If the user doesn't have permissions, but this doesn't prevent
        user from calling post with the params to perform the action.

        If the permissions are denied because of Permission in the database,
        Django would redirect to the changelist.

        It should also respect the has_xxx_permission methods
        """
        # Create a user without permissions for action
        user = User.objects.create_user(  # nosec
            username="user",
            email="user@email.org",
            password="pass",
            is_active=True,
            is_staff=True,
            is_superuser=False,
        )
        # Give user permissions to ShopAdmin change, add, view and delete
        for permission in Permission.objects.filter(
            codename__in=["change_shop", "view_shop", "add_shop", "delete_shop"]
        ):
            user.user_permissions.add(permission)

        self.client.force_login(user)

        # ShopAdmin has defined:
        #   def has_delete_permission(self, request, obj=None):
        #       return request.user.is_superuser

        post_params = {
            "action": ["show_message"],
            "select_across": ["0"],
            "index": ["0"],
            "_selected_action": ["3", "2", "1"],
        }
        response = self.client.post(
            reverse("admin:market_shop_changelist"),
            data=post_params,
            follow=True,  # Follow the redirect to get content
        )
        self.assertIsNotNone(response)
        self.assertEqual(response.status_code, 200)
        # Should not use confirmaiton page
        self.assertEqual(
            response.template_name,
            [
                "admin/market/shop/change_list.html",
                "admin/market/change_list.html",
                "admin/change_list.html",
            ],
        )

        # The action was to show user a message, and should not happen yet
        self.assertNotIn("You selected", response.rendered_content)

        # Django won't show the action as an option to you
        self.assertIn("No action selected", response.rendered_content)

    def test_no_permissions_in_code_superuser_for_action_with_confirmation(self):
        """
        Django would not show the action in changelist action selector
        If the user doesn't have permissions, but this doesn't prevent
        user from calling post with the params to perform the action.

        When permissions are denied from a change in code
        (ie has_xxx_permission in ModelAdmin), Django should still
        redirect to changelist. This should be true even if the user is
        a superuser.
        """
        # ShopAdmin has defined:
        #   def has_delete_permission(self, request, obj=None):
        #       return request.user.is_superuser

        ShopAdmin.has_delete_permission = lambda self, request, obj=None: False
        post_params = {
            "action": ["show_message"],
            "select_across": ["0"],
            "index": ["0"],
            "_selected_action": ["3", "2", "1"],
        }
        response = self.client.post(
            reverse("admin:market_shop_changelist"),
            data=post_params,
            follow=True,  # Follow the redirect to get content
        )
        self.assertIsNotNone(response)
        self.assertEqual(response.status_code, 200)
        # Should not use confirmaiton page
        self.assertEqual(
            response.template_name,
            [
                "admin/market/shop/change_list.html",
                "admin/market/change_list.html",
                "admin/change_list.html",
            ],
        )

        # The action was to show user a message, and should not happen yet
        self.assertNotIn("You selected", response.rendered_content)

        # Django won't show the action as an option to you
        self.assertIn("No action selected", response.rendered_content)

        # Remove our modification for ShopAdmin
        ShopAdmin.has_delete_permission = lambda self, request, obj=None: request.user.is_superuser

    def test_confirm_action_submit_button_should_perform_action(self):
        """
        The submit button should have param "_confirm_action"

        Simulate calling the post request that the button would
        """
        post_params = {
            CONFIRM_ACTION: ["Yes, I'm sure"],
            "action": ["show_message"],
            "_selected_action": ["3", "2", "1"],
        }
        response = self.client.post(
            reverse("admin:market_shop_changelist"),
            data=post_params,
            follow=True,  # Follow the redirect to get content
        )
        self.assertIsNotNone(response)
        self.assertEqual(response.status_code, 200)
        # Should not use confirmaiton page, since we clicked Yes, I'm sure
        self.assertEqual(
            response.template_name,
            [
                "admin/market/shop/change_list.html",
                "admin/market/change_list.html",
                "admin/change_list.html",
            ],
        )

        # The action was to show user a message, and should happen
        self.assertIn("You selected", response.rendered_content)

    def test_should_use_action_confirmation_template_if_set(self):
        expected_template = "market/admin/my_custom_template.html"
        ShopAdmin.action_confirmation_template = expected_template
        admin = ShopAdmin(Shop, AdminSite())
        actual_template = admin.render_action_confirmation(self.factory.request(), context={}).template_name
        self.assertEqual(expected_template, actual_template)
        # Clear our setting to not affect other tests
        ShopAdmin.action_confirmation_template = None

    def test_get_actions_when_empty(self):
        admin = InventoryAdmin(Inventory, AdminSite())
        admin.actions = None
        request = MagicMock()
        request.GET = {
            IS_POPUP_VAR,
        }
        actions = admin._get_actions({})
        self.assertEqual(actions, {})



from enum import Enum

from django.conf import settings

SAVE = "_save"
SAVE_AS_NEW = "_saveasnew"
ADD_ANOTHER = "_addanother"
SAVE_AND_CONTINUE = "_continue"
SAVE_ACTIONS = [SAVE, SAVE_AS_NEW, ADD_ANOTHER, SAVE_AND_CONTINUE]

CONFIRM_ADD = "_confirm_add"
CONFIRM_CHANGE = "_confirm_change"
CONFIRMATION_RECEIVED = "_confirmation_received"
CONFIRM_ACTION = "_confirm_action"
CONFIRM_FORM = "_form_action"
BACK = "_back"
CANCEL = "_cancel"

CACHE_TIMEOUT = getattr(settings, "ADMIN_CONFIRM_CACHE_TIMEOUT", 1000)
CACHE_KEYS = {
    "object": "admin_confirm__confirmation_object",
    "post": "admin_confirm__confirmation_request_post",
}
CACHE_KEY_PREFIX = getattr(settings, "ADMIN_CONFIRM_CACHE_KEY_PREFIX", "admin_confirm__file_cache")


DEBUG = getattr(settings, "ADMIN_CONFIRM_DEBUG", False)


FUNCTION_MARKER = "__finish_step__"


class ToolAction(Enum):
    CANCEL = "cancel"
    BACK = "back"
    CONFIRMED = "confirmed"
    FORWARD = "forward"
    INIT = "init"



"""
Tests confirmation of action that uses django-object-action
"""
from importlib import reload

from selenium.webdriver.common.by import By
from selenium.webdriver.remote.file_detector import LocalFileDetector

from admin_action_tools.constants import CONFIRM_ACTION
from admin_action_tools.tests.helpers import AdminConfirmIntegrationTestCase
from tests.factories import InventoryFactory, ShopFactory
from tests.market.admin import shoppingmall_admin


class ConfirmWithChangeActionTests(AdminConfirmIntegrationTestCase):
    def setUp(self):
        self.selenium.file_detector = LocalFileDetector()
        super().setUp()

    def tearDown(self):
        reload(shoppingmall_admin)
        super().tearDown()

    def test_change_and_changelist_action(self):
        shop = ShopFactory()
        inv1 = InventoryFactory(shop=shop, quantity=10)
        inv2 = InventoryFactory(shop=shop, quantity=10)

        self.selenium.get(self.live_server_url + f"/admin/market/inventory/{inv1.id}/actions/quantity_up")
        # Should ask for confirmation of action
        self.assertIn(CONFIRM_ACTION, self.selenium.page_source)

        self.selenium.find_element(By.NAME, CONFIRM_ACTION).click()
        inv1.refresh_from_db()
        self.assertEqual(inv1.quantity, 11)

        self.selenium.get(self.live_server_url + f"/admin/market/inventory/{inv2.id}/actions/quantity_up")
        # Should ask for confirmation of action
        self.assertIn(CONFIRM_ACTION, self.selenium.page_source)

        self.selenium.find_element(By.NAME, CONFIRM_ACTION).click()
        inv2.refresh_from_db()
        self.assertEqual(inv2.quantity, 11)

        self.selenium.get(self.live_server_url + "/admin/market/inventory/actions/quantity_down/")
        # Should ask for confirmation of action
        self.assertIn(CONFIRM_ACTION, self.selenium.page_source)
        self.selenium.find_element(By.NAME, CONFIRM_ACTION).click()

        inv1.refresh_from_db()
        inv2.refresh_from_db()
        self.assertEqual(inv1.quantity, 0)
        self.assertEqual(inv2.quantity, 0)



from admin_action_tools.admin.confirm_tool import AdminConfirmMixin, confirm_action
from admin_action_tools.admin.form_tool import ActionFormMixin, add_form_to_action

__all__ = [
    "AdminConfirmMixin",
    "confirm_action",
    "ActionFormMixin",
    "add_form_to_action",
]



"""
Tests confirmation of action that uses django-object-action
"""
from importlib import reload

from selenium.webdriver.common.by import By
from selenium.webdriver.remote.file_detector import LocalFileDetector

from admin_action_tools.constants import BACK, CANCEL, CONFIRM_ACTION, CONFIRM_FORM
from admin_action_tools.tests.helpers import AdminConfirmIntegrationTestCase
from tests.factories import InventoryFactory, ShopFactory
from tests.market.admin import shoppingmall_admin
from tests.market.form import NoteActionForm, NoteClearForm

CONFIRM_FORM_UNIQUE = f"{CONFIRM_FORM}_{NoteActionForm.__name__}"
CONFIRM_FORM_UNIQUE_2 = f"{CONFIRM_FORM}_{NoteClearForm.__name__}"


class MultipleFormActionTests(AdminConfirmIntegrationTestCase):
    def setUp(self):
        self.selenium.file_detector = LocalFileDetector()
        super().setUp()

    def tearDown(self):
        reload(shoppingmall_admin)
        super().tearDown()

    def test_change_form_and_confirm(self):
        shop = ShopFactory()
        inv1 = InventoryFactory(shop=shop, quantity=10)

        self.selenium.get(self.live_server_url + f"/admin/market/inventory/{inv1.id}/actions/add_notes_with_clear")
        # Should ask for confirmation of action
        self.assertIn(CONFIRM_FORM_UNIQUE, self.selenium.page_source)

        elems = self.selenium.find_elements(By.CLASS_NAME, "datetimeshortcuts")
        for elem in elems:
            elem.find_element(By.TAG_NAME, "a").click()

        self.selenium.find_element(By.ID, "id_note").send_keys("aaaaaa")

        self.selenium.find_element(By.NAME, CONFIRM_FORM_UNIQUE).click()

        # second form
        self.assertIn(CONFIRM_FORM_UNIQUE_2, self.selenium.page_source)
        self.selenium.find_element(By.ID, "id_clear_notes").click()
        self.selenium.find_element(By.NAME, CONFIRM_FORM_UNIQUE_2).click()

        # Should ask for confirmation of action
        self.assertIn(CONFIRM_ACTION, self.selenium.page_source)

        self.assertIn("aaaaaa", self.selenium.page_source)

        self.selenium.find_element(By.NAME, CONFIRM_ACTION).click()

        inv1.refresh_from_db()
        self.assertIn("aaaaaa", inv1.notes)

    def test_change_form_and_confirm_confirm_back(self):
        shop = ShopFactory()
        inv1 = InventoryFactory(shop=shop, quantity=10)

        self.selenium.get(self.live_server_url + f"/admin/market/inventory/{inv1.id}/actions/add_notes_with_clear")

        # Should ask for confirmation form
        self.assertIn(CONFIRM_FORM_UNIQUE, self.selenium.page_source)

        elems = self.selenium.find_elements(By.CLASS_NAME, "datetimeshortcuts")
        for elem in elems:
            elem.find_element(By.TAG_NAME, "a").click()

        self.selenium.find_element(By.ID, "id_note").send_keys("aaaaaa")

        self.selenium.find_element(By.NAME, CONFIRM_FORM_UNIQUE).click()

        # second form
        self.assertIn(CONFIRM_FORM_UNIQUE_2, self.selenium.page_source)
        self.selenium.find_element(By.ID, "id_clear_notes").click()
        self.selenium.find_element(By.NAME, CONFIRM_FORM_UNIQUE_2).click()

        # Should ask for confirmation of action
        self.assertIn(CONFIRM_ACTION, self.selenium.page_source)
        self.selenium.find_element(By.NAME, BACK).click()

        # Should ask for confirmation form
        self.assertIn(CONFIRM_FORM_UNIQUE_2, self.selenium.page_source)

        self.selenium.find_element(By.NAME, BACK).click()

        # Should ask for confirmation form
        self.assertIn(CONFIRM_FORM_UNIQUE, self.selenium.page_source)

        # back again
        self.selenium.find_element(By.CLASS_NAME, "cancel-link-nojs").click()
        self.assertTrue(self.selenium.current_url.endswith(f"/admin/market/inventory/{inv1.id}/change/"))



"""
Tests with different form input types
"""
from datetime import timedelta
from importlib import reload

from django.contrib.admin import VERTICAL
from django.utils import timezone
from selenium.webdriver.common.by import By

from admin_action_tools.constants import CONFIRM_ADD, CONFIRM_CHANGE
from admin_action_tools.tests.helpers import AdminConfirmIntegrationTestCase
from tests.factories import ShopFactory, TransactionFactory
from tests.market.admin import item_admin, shoppingmall_admin
from tests.market.models import GeneralManager, Item, ShoppingMall, Town


class ConfirmWithFormInputTypes(AdminConfirmIntegrationTestCase):
    @classmethod
    def setUpClass(cls):
        i_admin = item_admin.ItemAdmin
        i_admin.confirm_add = True
        i_admin.confirm_change = True
        i_admin.confirmation_fields = ["currency", "price", "name"]
        i_admin.radio_fields = {"currency": VERTICAL}

        mall_admin = shoppingmall_admin.ShoppingMallAdmin
        mall_admin.raw_id_fields = ["general_manager"]
        mall_admin.inlines = []
        return super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        reload(shoppingmall_admin)
        reload(item_admin)
        return super().tearDownClass()

    def test_radio_input_should_work(self):
        self.selenium.get(self.live_server_url + "/admin/market/item/add/")
        self.assertIn("radio", self.selenium.page_source)

        # Should ask for confirmation of add
        self.assertIn(CONFIRM_ADD, self.selenium.page_source)

        name = self.selenium.find_element(By.NAME, "name")
        name.send_keys("New Name")
        price = self.selenium.find_element(By.NAME, "price")
        price.send_keys("24.00")

        # Select a radio option for currency
        currency = self.selenium.find_element(By.ID, "id_currency_0")
        currency.click()
        expected_value = currency.get_attribute("value")

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should have hidden form containing the updated name
        self.assertIn("Confirm", self.selenium.page_source)
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        name = hidden_form.find_element(By.NAME, "name")
        self.assertEqual("New Name", name.get_attribute("value"))

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should persist addition
        self.assertEqual(Item.objects.count(), 1)
        item = Item.objects.all().first()
        self.assertEqual("New Name", item.name)
        self.assertEqual(expected_value, item.currency)
        self.assertEqual(24, int(item.price))

    def test_raw_id_fields_should_work(self):
        gm1 = GeneralManager.objects.create(name="gm1")
        shops = [ShopFactory(name=i) for i in range(3)]
        town = Town.objects.create(name="town")
        mall = ShoppingMall.objects.create(name="mall", general_manager=gm1, town=town)
        mall.shops.set(shops)

        self.selenium.get(self.live_server_url + f"/admin/market/shoppingmall/{mall.id}/change/")
        self.assertIn(CONFIRM_CHANGE, self.selenium.page_source)

        # Make a change to trigger confirmation page
        name = self.selenium.find_element(By.NAME, "name")
        name.send_keys("New Name")

        # Set general_manager via raw_id_fields
        gm2 = GeneralManager.objects.create(name="gm2")
        general_manager = self.selenium.find_element(By.NAME, "general_manager")
        general_manager.clear()
        general_manager.send_keys(str(gm2.id))

        self.selenium.find_element(By.NAME, "_continue").click()

        self.assertIn("Confirm", self.selenium.page_source)
        self.selenium.find_element(By.NAME, "_continue").click()

        mall.refresh_from_db()
        self.assertIn("New Name", mall.name)
        self.assertEqual(gm2, mall.general_manager)

    def test_datetime_and_field_should_work(self):
        original_timestamp = timezone.now() - timedelta(hours=1)
        transaction = TransactionFactory(timestamp=original_timestamp)

        self.selenium.get(self.live_server_url + f"/admin/market/transaction/{transaction.id}/change/")
        self.assertIn(CONFIRM_CHANGE, self.selenium.page_source)

        # Set date via text input
        date_input = self.selenium.find_element(By.ID, "id_date")
        date_input.clear()
        date_input.send_keys("2021-01-01")
        self.assertEqual(date_input.get_attribute("value"), "2021-01-01")

        # Set timestamp via text input
        timestamp_date = self.selenium.find_element(By.ID, "id_timestamp_0")
        timestamp_date.clear()
        timestamp_date.send_keys(str(timezone.now().date()))
        timestamp_time = self.selenium.find_element(By.ID, "id_timestamp_1")
        timestamp_time.clear()
        timestamp_time.send_keys(str(timezone.now().time()))

        # Click save and continue
        self.selenium.find_element(By.NAME, "_continue").click()

        # Click Yes I'm Sure on confirmation page
        self.assertIn("Confirm", self.selenium.page_source)
        self.selenium.find_element(By.NAME, "_continue").click()

        transaction.refresh_from_db()
        self.assertEqual(str(transaction.date), "2021-01-01")
        self.assertTrue(transaction.timestamp > original_timestamp)



from __future__ import annotations

import functools
from datetime import datetime, timedelta
from typing import Dict, Optional, Tuple

from django.http import HttpRequest, QueryDict

from admin_action_tools.constants import BACK, CANCEL, FUNCTION_MARKER, ToolAction


def gather_tools(func):
    """
    @gather_tools function is a wrapper that is automatically added.
    It allows django-admin-action-tools to finalize the processing.
    """

    @functools.wraps(func)
    def func_wrapper(modeladmin, request, queryset_or_object):

        tool_chain: ToolChain = ToolChain(request)
        # get result
        forms = modeladmin.get_tools_result(tool_chain)
        kwargs = {}
        if len(forms) == 1:
            kwargs["form"] = forms[0]
        elif len(forms) > 1:
            kwargs["forms"] = forms

        # clear session
        tool_chain.clear_tool_chain()

        return func(modeladmin, request, queryset_or_object, **kwargs)

    return func_wrapper


def add_finishing_step(func):
    if not hasattr(func, FUNCTION_MARKER):
        setattr(func, FUNCTION_MARKER, True)
        return gather_tools(func)
    return func


class ToolChain:
    def __init__(self, request: HttpRequest) -> None:
        self.request = request
        self.session = request.session
        self.name = f"toolchain{request.path}"
        self._get_data()
        self.data.setdefault("history", [])

    def _get_data(self):
        old_data = self.session.get(self.name, {})
        expire_at = old_data.get("expire_at")

        if expire_at:
            try:
                expire_at = datetime.fromisoformat(expire_at)
            except Exception:  # pylint: disable=broad-except
                expire_at = None
                old_data = None

        if not old_data:
            self.data = {"expire_at": self._get_expiration()}
            self._save()
        elif expire_at and expire_at < datetime.now():
            self.data = {"expire_at": self._get_expiration()}
            self._save()
        else:
            self.data = old_data

    def _update_expire_at(self):
        self.data["expire_at"] = self._get_expiration()
        self._save()

    @staticmethod
    def _get_expiration():
        return (datetime.now() + timedelta(seconds=60)).isoformat()

    def _save(self):
        self.session[self.name] = self.data
        self.session.modified = True

    def get_toolchain(self) -> Dict:
        return self.data

    def set_tool(self, tool_name: str, data: dict, metadata=None) -> None:
        self.data["history"].append(tool_name)
        cleaned_data = self.__clean_data(data, metadata)
        self.data.update({tool_name: cleaned_data})
        self._save()

    def get_tool(self, tool_name: str) -> Tuple[Optional[dict], Optional[dict]]:
        tool = self.data.get(tool_name, {})
        return QueryDict(tool.get("data")), tool.get("metadata")

    def clear_tool_chain(self):
        self.session.pop(self.name, None)

    def is_rollback(self):
        return BACK in self.request.POST

    def is_cancel(self):
        return CANCEL in self.request.POST

    def rollback(self):
        tool_name = self.data["history"].pop()
        data, _ = self.get_tool(tool_name)
        del self.data[tool_name]
        self._save()
        return data

    def is_first_tool(self):
        return not self.data["history"]

    def get_next_step(self, tool_name: str) -> ToolAction:
        self._update_expire_at()
        if self.is_cancel():
            return ToolAction.CANCEL
        if self.is_rollback():
            tool_to_rollback = self.get_history()[-1]
            if tool_to_rollback != tool_name:
                return ToolAction.FORWARD
            return ToolAction.BACK
        if tool_name in self.request.POST:
            return ToolAction.CONFIRMED
        if tool_name in self.get_toolchain():
            return ToolAction.FORWARD
        return ToolAction.INIT

    def __clean_data(self, data: QueryDict, metadata):
        new_data = data.copy()
        new_data.pop("csrfmiddlewaretoken", None)

        metadata = metadata or {}
        return {"data": new_data.urlencode(), "metadata": metadata}

    def get_history(self):
        return self.data["history"]



from unittest import mock

from django.urls import reverse

from admin_action_tools.tests.helpers import AdminConfirmTestCase
from tests.factories import ShopFactory
from tests.market.admin import ShoppingMallAdmin
from tests.market.models import ShoppingMall


@mock.patch.object(ShoppingMallAdmin, "inlines", [])
class TestConfirmChangeAndAddM2MField(AdminConfirmTestCase):
    def test_post_add_without_confirm_add_m2m(self):
        shops = [ShopFactory() for i in range(3)]

        data = {"name": "name", "shops": [s.id for s in shops]}
        response = self.client.post(reverse("admin:market_shoppingmall_add"), data)
        # Redirects to changelist and is added
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/market/shoppingmall/")
        self.assertEqual(ShoppingMall.objects.count(), 1)
        self.assertEqual(ShoppingMall.objects.all().first().shops.count(), 3)

    def test_post_add_with_confirm_add_m2m(self):
        ShoppingMallAdmin.confirmation_fields = ["shops"]
        shops = [ShopFactory() for i in range(3)]

        data = {
            "name": "name",
            "shops": [s.id for s in shops],
            "_confirm_add": True,
            "_save": True,
        }
        response = self.client.post(reverse("admin:market_shoppingmall_add"), data)

        # Ensure not redirected (confirmation page does not redirect)
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/shoppingmall/confirm_tool/change_confirmation.html",
            "admin/market/confirm_tool/change_confirmation.html",
            "admin/confirm_tool/change_confirmation.html",
        ]
        self.assertEqual(response.template_name, expected_templates)

        self._assertManyToManyFormHtml(
            rendered_content=response.rendered_content,
            options=shops,
            selected_ids=data["shops"],
        )
        self._assertSubmitHtml(rendered_content=response.rendered_content)

        # Should not have been added yet
        self.assertEqual(ShoppingMall.objects.count(), 0)

        # Confirmation page would not have the _confirm_add sent on submit
        del data["_confirm_add"]
        # Selecting to "Yes, I'm sure" on the confirmation page
        # Would post to the same endpoint
        response = self.client.post(reverse("admin:market_shoppingmall_add"), data)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/market/shoppingmall/")
        self.assertEqual(ShoppingMall.objects.count(), 1)
        self.assertEqual(ShoppingMall.objects.all().first().shops.count(), 3)

    def test_m2m_field_post_change_with_confirm_change(self):
        shops = [ShopFactory() for i in range(10)]
        shopping_mall = ShoppingMall.objects.create(name="My Mall")
        shopping_mall.shops.set(shops)
        # Currently ShoppingMall configured with confirmation_fields = ['name']
        data = {
            "name": "Not My Mall",
            "shops": [1, 2],
            "id": shopping_mall.id,
            "_confirm_change": True,
            "csrfmiddlewaretoken": "fake token",
            "_continue": True,
        }
        response = self.client.post(f"/admin/market/shoppingmall/{shopping_mall.id}/change/", data)
        # Ensure not redirected (confirmation page does not redirect)
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/shoppingmall/confirm_tool/change_confirmation.html",
            "admin/market/confirm_tool/change_confirmation.html",
            "admin/confirm_tool/change_confirmation.html",
        ]
        self.assertEqual(response.template_name, expected_templates)

        # Should show two lists for the m2m current and modified values
        self.assertEqual(response.rendered_content.count("<ul>"), 2)

        self._assertManyToManyFormHtml(
            rendered_content=response.rendered_content,
            options=shops,
            selected_ids=data["shops"],
        )
        self._assertSubmitHtml(rendered_content=response.rendered_content, save_action="_continue")

        # Hasn't changed item yet
        shopping_mall.refresh_from_db()
        self.assertEqual(shopping_mall.name, "My Mall")

        # Selecting to "Yes, I'm sure" on the confirmation page
        # Would post to the same endpoint
        del data["_confirm_change"]
        response = self.client.post(f"/admin/market/shoppingmall/{shopping_mall.id}/change/", data)
        # will show the change page for this shopping_mall
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, f"/admin/market/shoppingmall/{shopping_mall.id}/change/")
        # Should not be the confirmation page, we already confirmed change
        self.assertNotEqual(response.templates, expected_templates)
        self.assertEqual(ShoppingMall.objects.count(), 1)
        self.assertEqual(ShoppingMall.objects.all().first().shops.count(), 2)

    def test_m2m_field_post_change_with_confirm_change_multiple_selected(self):
        shops = [ShopFactory() for i in range(10)]
        shopping_mall = ShoppingMall.objects.create(name="My Mall")
        shopping_mall.shops.set(shops)
        # Currently ShoppingMall configured with confirmation_fields = ['name']
        data = {
            "name": "Not My Mall",
            "shops": [1, 2, 3],
            "id": shopping_mall.id,
            "_confirm_change": True,
            "csrfmiddlewaretoken": "fake token",
            "_save": True,
        }
        response = self.client.post(f"/admin/market/shoppingmall/{shopping_mall.id}/change/", data)
        # Ensure not redirected (confirmation page does not redirect)
        self.assertEqual(response.status_code, 200)
        expected_templates = [
            "admin/market/shoppingmall/confirm_tool/change_confirmation.html",
            "admin/market/confirm_tool/change_confirmation.html",
            "admin/confirm_tool/change_confirmation.html",
        ]
        self.assertEqual(response.template_name, expected_templates)

        # Hasn't changed item yet
        shopping_mall.refresh_from_db()
        self.assertEqual(shopping_mall.name, "My Mall")

    def test_post_change_without_confirm_change_m2m_value(self):
        # make the m2m the confirmation_field
        ShoppingMallAdmin.confirmation_fields = ["shops"]
        shops = [ShopFactory() for i in range(3)]
        shopping_mall = ShoppingMall.objects.create(name="name")
        shopping_mall.shops.set(shops)
        assert shopping_mall.shops.count() == 3  # nosec

        data = {"name": "name", "id": str(shopping_mall.id), "shops": ["1"]}
        response = self.client.post(f"/admin/market/shoppingmall/{shopping_mall.id}/change/", data)
        # Redirects to changelist
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/market/shoppingmall/")
        # Shop has changed
        shopping_mall.refresh_from_db()
        self.assertEqual(shopping_mall.shops.count(), 1)

    def test_form_invalid_m2m_value(self):
        ShoppingMallAdmin.confirmation_fields = ["shops"]
        shopping_mall = ShoppingMall.objects.create(name="name")

        data = {
            "id": shopping_mall.id,
            "name": "name",
            "shops": [1, 2, 3],  # These shops don't exist
            "_confirm_change": True,
            "csrfmiddlewaretoken": "fake token",
        }
        response = self.client.post(f"/admin/market/shoppingmall/{shopping_mall.id}/change/", data)

        # Form invalid should show errors on form
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(response.context_data.get("errors"))
        self.assertTrue(str(response.context_data["errors"][0][0]).startswith("Select a valid choice."))
        # Should not have updated inventory
        shopping_mall.refresh_from_db()
        self.assertEqual(shopping_mall.shops.count(), 0)



"""
Ensures that confirmations work with validators on the Model and on the Modelform.
"""

from importlib import reload
from unittest import mock

from selenium.webdriver.common.by import By

from admin_action_tools.constants import CONFIRM_ADD
from admin_action_tools.tests.helpers import AdminConfirmIntegrationTestCase
from tests.factories import (
    InventoryFactory,
    ItemFactory,
    ShopFactory,
    TransactionFactory,
)
from tests.market.admin import shoppingmall_admin
from tests.market.models import ItemSale


class ConfirmWithValidatorsTests(AdminConfirmIntegrationTestCase):
    def setUp(self):
        self.admin = shoppingmall_admin.ShoppingMallAdmin
        self.admin.inlines = [shoppingmall_admin.ShopInline]
        super().setUp()

    def tearDown(self):
        reload(shoppingmall_admin)
        super().tearDown()

    @mock.patch("tests.market.models.ItemSale.clean")
    def test_can_confirm_for_models_with_validator_on_model_field(self, _mock_clean):
        # ItemSale.currency has a validator on it
        ItemFactory()
        TransactionFactory()

        self.selenium.get(self.live_server_url + "/admin/market/itemsale/add/")
        # Should ask for confirmation of change
        self.assertIn(CONFIRM_ADD, self.selenium.page_source)

        self.set_value(by=By.NAME, by_value="quantity", value="1")
        self.set_value(by=By.NAME, by_value="total", value="10.00")
        self.set_value(by=By.NAME, by_value="currency", value="USD")
        self.select_index(by=By.NAME, by_value="transaction", index=1)
        self.select_index(by=By.NAME, by_value="item", index=1)

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should have hidden form containing the updated name
        self.assertIn("Confirm", self.selenium.page_source)
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        currency = hidden_form.find_element(By.NAME, "currency")
        self.assertIn("USD", currency.get_attribute("value"))

        # Should not have been added yet
        self.assertEqual(ItemSale.objects.count(), 0)
        self.selenium.find_element(By.NAME, "_continue").click()

        # Should persist change
        self.assertEqual(ItemSale.objects.count(), 1)

    def test_cannot_confirm_for_models_with_validator_on_model_field_if_validator_fails(
        self,
    ):
        # ItemSale.currency has a validator on it
        shop = ShopFactory()
        item = ItemFactory()
        InventoryFactory(shop=shop, item=item, quantity=10)
        TransactionFactory(shop=shop)

        self.selenium.get(self.live_server_url + "/admin/market/itemsale/add/")
        # Should ask for confirmation of change
        self.assertIn(CONFIRM_ADD, self.selenium.page_source)

        self.set_value(by=By.NAME, by_value="quantity", value="1")
        self.set_value(by=By.NAME, by_value="total", value="10.00")
        self.set_value(by=By.NAME, by_value="currency", value="FAKE")
        self.select_index(by=By.NAME, by_value="transaction", index=1)
        self.select_index(by=By.NAME, by_value="item", index=1)

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should show errors and not confirmation page
        self.assertNotIn("Confirm", self.selenium.page_source)
        self.assertIn("Invalid Currency", self.selenium.page_source)
        self.assertIn(CONFIRM_ADD, self.selenium.page_source)

        # Should not have been added yet
        self.assertEqual(ItemSale.objects.count(), 0)

        # Now fix the issue
        self.set_value(by=By.NAME, by_value="currency", value="USD")
        self.selenium.find_element(By.NAME, "_continue").click()

        # Should have hidden form containing the updated currency
        self.assertIn("Confirm", self.selenium.page_source)
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        currency = hidden_form.find_element(By.NAME, "currency")
        self.assertIn("USD", currency.get_attribute("value"))

        # Should not have been added yet
        self.assertEqual(ItemSale.objects.count(), 0)
        self.selenium.find_element(By.NAME, "_continue").click()

        # Should persist change
        self.assertEqual(ItemSale.objects.count(), 1)

    def test_can_confirm_for_models_with_clean_overridden(self):
        shop = ShopFactory()
        item = ItemFactory()
        InventoryFactory(shop=shop, item=item, quantity=10)
        transaction = TransactionFactory(shop=shop)

        self.selenium.get(self.live_server_url + "/admin/market/itemsale/add/")
        # Should ask for confirmation of change
        self.assertIn(CONFIRM_ADD, self.selenium.page_source)

        self.set_value(by=By.NAME, by_value="quantity", value="9")
        self.set_value(by=By.NAME, by_value="total", value="10.00")
        self.set_value(by=By.NAME, by_value="currency", value="USD")
        self.select_index(by=By.NAME, by_value="transaction", index=1)
        self.select_index(by=By.NAME, by_value="item", index=1)

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should not have been added yet
        self.assertEqual(ItemSale.objects.count(), 0)

        # Should have hidden form containing the updated currency
        self.assertIn("Confirm", self.selenium.page_source)
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        currency = hidden_form.find_element(By.NAME, "currency")
        self.assertIn("USD", currency.get_attribute("value"))

        # Confirm the change
        self.selenium.find_element(By.NAME, "_continue").click()

        # Should persist change
        self.assertEqual(ItemSale.objects.count(), 1)

        # Ensure that the date and timestamp saved correctly
        item_sale = ItemSale.objects.first()
        self.assertEqual(item_sale.transaction, transaction)
        self.assertEqual(item_sale.item, item)
        self.assertEqual(item_sale.currency, "USD")

    def test_cannot_confirm_for_models_with_clean_overridden_if_clean_fails(self):
        shop = ShopFactory()
        item = ItemFactory()
        InventoryFactory(shop=shop, item=item, quantity=1)
        TransactionFactory(shop=shop)

        self.selenium.get(self.live_server_url + "/admin/market/itemsale/add/")
        # Should ask for confirmation of change
        self.assertIn(CONFIRM_ADD, self.selenium.page_source)

        self.set_value(by=By.NAME, by_value="quantity", value="9")
        self.set_value(by=By.NAME, by_value="total", value="10.00")
        self.set_value(by=By.NAME, by_value="currency", value="USD")
        self.select_index(by=By.NAME, by_value="transaction", index=1)
        self.select_index(by=By.NAME, by_value="item", index=1)

        self.selenium.find_element(By.NAME, "_continue").click()

        # Should show errors and not confirmation page
        self.assertNotIn("Confirm", self.selenium.page_source)
        self.assertIn("Shop does not have enough of the item stocked", self.selenium.page_source)
        self.assertIn(CONFIRM_ADD, self.selenium.page_source)

        # Should not have been added yet
        self.assertEqual(ItemSale.objects.count(), 0)

        # Now fix the issue
        self.set_value(by=By.NAME, by_value="quantity", value="1")
        self.selenium.find_element(By.NAME, "_continue").click()

        # Should have hidden form containing the updated currency
        self.assertIn("Confirm", self.selenium.page_source)
        hidden_form = self.selenium.find_element(By.ID, "hidden-form")
        quantity = hidden_form.find_element(By.NAME, "quantity")
        self.assertIn("1", str(quantity.get_attribute("value")))

        # Confirm change
        self.selenium.find_element(By.NAME, "_continue").click()

        # Should persist change
        self.assertEqual(ItemSale.objects.count(), 1)



""" FileCache - caches files for ModelAdmins with confirmations.

Code modified from: https://github.com/MaistrenkoAnton/filefield-cache/blob/master/filefield_cache/cache.py
Original copy date: April 22, 2021
---
Modified to be compatible with more versions of Django/Python
---
MIT License

Copyright (c) 2020 Maistrenko Anton

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from django.core.files.uploadedfile import InMemoryUploadedFile

try:
    from cStringIO import StringIO as BytesIO  # noqa: WPS433
except ImportError:
    from io import BytesIO  # noqa: WPS433, WPS440

from django.core.cache import cache

from admin_action_tools.constants import CACHE_TIMEOUT
from admin_action_tools.utils import log


class FileCache(object):
    "Cache file data and retain the file upon confirmation."

    timeout = CACHE_TIMEOUT

    def __init__(self):
        self.cache = cache
        self.cached_keys = []

    def set(self, key, upload):
        """
        Set file data to cache for 1000s

        :param key: cache key
        :param upload: file data
        """
        try:  # noqa: WPS229
            state = {
                "name": upload.name,
                "size": upload.size,
                "content_type": upload.content_type,
                "charset": upload.charset,
                "content": upload.file.read(),
            }
            upload.file.seek(0)
            self.cache.set(key, state, self.timeout)
            log(f"Setting file cache with {key}")
            self.cached_keys.append(key)
        except AttributeError:  # pragma: no cover
            pass  # noqa: WPS420

    def get(self, key):
        """
        Get the file data from cache using specific cache key

        :param key: cache key
        :return: File data
        """
        upload = None
        state = self.cache.get(key)
        if state:
            file = BytesIO()
            file.write(state["content"])
            upload = InMemoryUploadedFile(
                file=file,
                field_name="file",
                name=state["name"],
                content_type=state["content_type"],
                size=state["size"],
                charset=state["charset"],
            )
            upload.file.seek(0)
            log(f"Getting file cache with {key}")
        return upload

    def delete(self, key):
        """
        Delete file data from cache

        :param key: cache key
        """
        self.cache.delete(key)
        self.cached_keys.remove(key)

    def delete_all(self):
        "Delete all cached file data from cache."
        if self.cached_keys:
            self.cache.delete_many(self.cached_keys)
            self.cached_keys = []



from unittest import mock

from django.core.cache import cache

from admin_action_tools.constants import CACHE_KEYS
from admin_action_tools.tests.helpers import AdminConfirmTestCase
from tests.factories import ShopFactory
from tests.market.admin import ShoppingMallAdmin
from tests.market.models import GeneralManager, ShoppingMall, Town


@mock.patch.object(ShoppingMallAdmin, "inlines", [])
class TestAdminOptions(AdminConfirmTestCase):
    @mock.patch.object(ShoppingMallAdmin, "confirmation_fields", ["name"])
    @mock.patch.object(ShoppingMallAdmin, "fields", ["name", "town"])
    def test_change_model_with_m2m_field_without_input_for_m2m_field_should_work(self):
        gm = GeneralManager.objects.create(name="gm")
        shops = [ShopFactory() for i in range(3)]
        town = Town.objects.create(name="town")
        mall = ShoppingMall.objects.create(name="mall", general_manager=gm, town=town)
        mall.shops.set(shops)

        # new values
        town2 = Town.objects.create(name="town2")

        data = {
            "id": mall.id,
            "name": "name",
            "town": town2.id,
            "_confirm_change": True,
            "_continue": True,
        }
        response = self.client.post(f"/admin/market/shoppingmall/{mall.id}/change/", data=data)

        # Should be shown confirmation page
        self._assertSubmitHtml(rendered_content=response.rendered_content, save_action="_continue")

        # Should not have cached the unsaved obj
        cached_item = cache.get(CACHE_KEYS["object"])
        self.assertIsNone(cached_item)

        # Should not have saved changes yet
        self.assertEqual(ShoppingMall.objects.count(), 1)
        mall.refresh_from_db()
        self.assertEqual(mall.name, "mall")
        self.assertEqual(mall.general_manager, gm)
        self.assertEqual(mall.town, town)
        for shop in mall.shops.all():
            self.assertIn(shop, shops)

        # Click "Yes, I'm Sure"
        confirmation_received_data = data
        del confirmation_received_data["_confirm_change"]

        response = self.client.post(
            f"/admin/market/shoppingmall/{mall.id}/change/",
            data=confirmation_received_data,
        )

        # Should not have redirected to changelist
        self.assertEqual(response.url, f"/admin/market/shoppingmall/{mall.id}/change/")

        # Should have saved obj
        self.assertEqual(ShoppingMall.objects.count(), 1)
        saved_item = ShoppingMall.objects.all().first()
        # should have updated fields that were in form
        self.assertEqual(saved_item.name, data["name"])
        self.assertEqual(saved_item.town, town2)
        # should have presevered the fields that are not in form
        self.assertEqual(saved_item.general_manager, gm)
        for shop in saved_item.shops.all():
            self.assertIn(shop, shops)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    @mock.patch.object(ShoppingMallAdmin, "confirmation_fields", ["name"])
    @mock.patch.object(ShoppingMallAdmin, "exclude", ["shops"])
    def test_when_m2m_field_in_exclude_changes_to_field_should_not_be_saved(self):
        gm = GeneralManager.objects.create(name="gm")
        shops = [ShopFactory() for i in range(3)]
        town = Town.objects.create(name="town")
        mall = ShoppingMall.objects.create(name="mall", general_manager=gm, town=town)
        mall.shops.set(shops)

        # new values
        gm2 = GeneralManager.objects.create(name="gm2")
        town2 = Town.objects.create(name="town2")

        data = {
            "id": mall.id,
            "name": "name",
            "general_manager": gm2.id,
            "shops": [1],
            "town": town2.id,
            "_confirm_change": True,
            "_continue": True,
        }
        response = self.client.post(f"/admin/market/shoppingmall/{mall.id}/change/", data=data)
        # Should be shown confirmation page
        self._assertSubmitHtml(rendered_content=response.rendered_content, save_action="_continue")

        # Should not have cached the unsaved obj
        cached_item = cache.get(CACHE_KEYS["object"])
        self.assertIsNone(cached_item)

        # Should not have saved changes yet
        self.assertEqual(ShoppingMall.objects.count(), 1)
        mall.refresh_from_db()
        self.assertEqual(mall.name, "mall")
        self.assertEqual(mall.general_manager, gm)
        self.assertEqual(mall.town, town)
        for shop in mall.shops.all():
            self.assertIn(shop, shops)

        # Click "Yes, I'm Sure"
        confirmation_received_data = data
        del confirmation_received_data["_confirm_change"]

        response = self.client.post(
            f"/admin/market/shoppingmall/{mall.id}/change/",
            data=confirmation_received_data,
        )

        # Should not have redirected to changelist
        self.assertEqual(response.url, f"/admin/market/shoppingmall/{mall.id}/change/")

        # Should have saved obj
        self.assertEqual(ShoppingMall.objects.count(), 1)
        saved_item = ShoppingMall.objects.all().first()
        # should have updated fields that were in form
        self.assertEqual(saved_item.name, data["name"])
        self.assertEqual(saved_item.town, town2)
        self.assertEqual(saved_item.general_manager, gm2)
        # should have presevered the fields that are not in form (exclude)
        for shop in saved_item.shops.all():
            self.assertIn(shop, shops)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))

    @mock.patch.object(ShoppingMallAdmin, "confirmation_fields", ["name"])
    @mock.patch.object(ShoppingMallAdmin, "exclude", ["shops", "name"])
    @mock.patch.object(ShoppingMallAdmin, "inlines", [])
    def test_if_confirmation_fields_in_exclude_should_not_trigger_confirmation(self):
        gm = GeneralManager.objects.create(name="gm")
        shops = [ShopFactory() for i in range(3)]
        town = Town.objects.create(name="town")
        mall = ShoppingMall.objects.create(name="mall", general_manager=gm, town=town)
        mall.shops.set(shops)

        # new values
        gm2 = GeneralManager.objects.create(name="gm2")
        town2 = Town.objects.create(name="town2")

        data = {
            "id": mall.id,
            "name": "name",
            "general_manager": gm2.id,
            "shops": [1],
            "town": town2.id,
            "_confirm_change": True,
            "_continue": True,
        }
        response = self.client.post(f"/admin/market/shoppingmall/{mall.id}/change/", data=data)
        # Should not be shown confirmation page
        # SInce we used "Save and Continue", should show change page
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, f"/admin/market/shoppingmall/{mall.id}/change/")

        # Should have saved the non excluded fields
        mall.refresh_from_db()
        for shop in shops:
            self.assertIn(shop, mall.shops.all())
        self.assertEqual(mall.name, "mall")
        # Should have saved other fields
        self.assertEqual(mall.town, town2)
        self.assertEqual(mall.general_manager, gm2)

    @mock.patch.object(ShoppingMallAdmin, "confirmation_fields", ["name"])
    @mock.patch.object(ShoppingMallAdmin, "readonly_fields", ["shops", "name"])
    @mock.patch.object(ShoppingMallAdmin, "inlines", [])
    def test_if_confirmation_fields_in_readonly_should_not_trigger_confirmation(self):
        gm = GeneralManager.objects.create(name="gm")
        shops = [ShopFactory() for i in range(3)]
        town = Town.objects.create(name="town")
        mall = ShoppingMall.objects.create(name="mall", general_manager=gm, town=town)
        mall.shops.set(shops)

        # new values
        gm2 = GeneralManager.objects.create(name="gm2")
        town2 = Town.objects.create(name="town2")

        data = {
            "id": mall.id,
            "name": "name",
            "general_manager": gm2.id,
            "shops": [1],
            "town": town2.id,
            "_confirm_change": True,
            "_continue": True,
        }
        response = self.client.post(f"/admin/market/shoppingmall/{mall.id}/change/", data=data)
        # Should not be shown confirmation page
        # SInce we used "Save and Continue", should show change page
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, f"/admin/market/shoppingmall/{mall.id}/change/")

        # Should have saved the non excluded fields
        mall.refresh_from_db()
        for shop in shops:
            self.assertIn(shop, mall.shops.all())
        self.assertEqual(mall.name, "mall")
        # Should have saved other fields
        self.assertEqual(mall.town, town2)
        self.assertEqual(mall.general_manager, gm2)

    @mock.patch.object(ShoppingMallAdmin, "confirmation_fields", ["name"])
    @mock.patch.object(ShoppingMallAdmin, "readonly_fields", ["shops"])
    @mock.patch.object(ShoppingMallAdmin, "inlines", [])
    def test_readonly_fields_should_not_change(self):
        gm = GeneralManager.objects.create(name="gm")
        shops = [ShopFactory() for i in range(3)]
        town = Town.objects.create(name="town")
        mall = ShoppingMall.objects.create(name="mall", general_manager=gm, town=town)
        mall.shops.set(shops)

        # new values
        gm2 = GeneralManager.objects.create(name="gm2")
        town2 = Town.objects.create(name="town2")

        data = {
            "id": mall.id,
            "name": "name",
            "general_manager": gm2.id,
            "shops": [1],
            "town": town2.id,
            "_confirm_change": True,
            "_continue": True,
        }
        response = self.client.post(f"/admin/market/shoppingmall/{mall.id}/change/", data=data)
        # Should be shown confirmation page
        self._assertSubmitHtml(rendered_content=response.rendered_content, save_action="_continue")

        # Should not have cached the unsaved obj
        cached_item = cache.get(CACHE_KEYS["object"])
        self.assertIsNone(cached_item)

        # Should not have saved changes yet
        self.assertEqual(ShoppingMall.objects.count(), 1)
        mall.refresh_from_db()
        self.assertEqual(mall.name, "mall")
        self.assertEqual(mall.general_manager, gm)
        self.assertEqual(mall.town, town)
        for shop in mall.shops.all():
            self.assertIn(shop, shops)

        # Click "Yes, I'm Sure"
        confirmation_received_data = data
        del confirmation_received_data["_confirm_change"]

        response = self.client.post(
            f"/admin/market/shoppingmall/{mall.id}/change/",
            data=confirmation_received_data,
        )

        # Should not have redirected to changelist
        self.assertEqual(response.url, f"/admin/market/shoppingmall/{mall.id}/change/")

        # Should have saved obj
        self.assertEqual(ShoppingMall.objects.count(), 1)
        saved_item = ShoppingMall.objects.all().first()
        # should have updated fields that were in form
        self.assertEqual(saved_item.name, data["name"])
        self.assertEqual(saved_item.town, town2)
        self.assertEqual(saved_item.general_manager, gm2)
        # should have presevered the fields that are not in form (exclude)
        for shop in saved_item.shops.all():
            self.assertIn(shop, shops)

        # Should have cleared cache
        for key in CACHE_KEYS.values():
            self.assertIsNone(cache.get(key))
